package eu.devity.vaultconnector.controller

import eu.devity.vaultconnector.serializable.*
import eu.devity.vaultconnector.serializable.vault.*
import org.springframework.http.HttpEntity
import org.springframework.http.HttpMethod
import org.springframework.http.HttpStatus
import org.springframework.vault.VaultException
import org.springframework.vault.core.VaultTemplate
import org.springframework.vault.support.VaultMount
import org.springframework.web.bind.annotation.*
import org.springframework.web.client.HttpClientErrorException
import org.springframework.web.util.UriUtils

// 10 years validity for root CA
const val ROOT_TTL = "87600h"

// 5 years validity for intermediate CA
const val INTERMEDIATE_TTL = "43800h"

@RestController
@RequestMapping("/cas")
class CAController(
    val vaultTemplate: VaultTemplate,
    val pkiManagedPath: String
) {

    @PostMapping("{caId}")
    @ResponseStatus(HttpStatus.CREATED)
    fun createCA(
        @PathVariable caId: String,
        @RequestBody request: CreateCARequest
    ): CreateCAResponse {
        val pkiPath = UriUtils.decode(caId, "ASCII")
        return if (request.caType == CAType.EXTERNAL) {
            CreateCAResponse(
                createIntermediateCsr(pkiPath, request).csr,
                null,
                "$pkiPath-int-01",
                pkiManagedPath
            )
        } else {
            createRootCA(pkiPath, request)
            val intermediateCsr = createIntermediateCsr(pkiPath, request).csr
            val data = signIntermediateCA(
                pkiPath,
                intermediateCsr,
                request.intermediateTTL ?: INTERMEDIATE_TTL
            )
            CreateCAResponse(
                null,
                "${data.certificate}\n${data.issuingCa}",
                "$pkiPath-int-01",
                pkiManagedPath
            )
        }
    }

    @PutMapping("{caId}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    fun setSignedCA(
        @PathVariable caId: String,
        @RequestBody request: SetSignedCARequest
    ) {
        val pkiPath = UriUtils.decode(caId, "ASCII")
        setSignedIntermediateCA(pkiPath, request.certificateChain)
    }

    fun createRootCA(
        caId: String,
        request: CreateCARequest
    ): VaultCreateCAResponse.Data {
        val csrFields = request.csrFields ?: CsrRequest()
        val builder = VaultMount.builder()
        val ttl = request.rootTTl ?: ROOT_TTL

        builder.type("pki")
        builder.config(
            mapOf(
                "max_lease_ttl" to ttl,
                "default_lease_ttl" to ttl
            )
        )
        builder.description("Root CA for $caId")

        val mount = builder.build()
        val sysOperations = vaultTemplate.opsForSys()
        val pkiPath = "$pkiManagedPath/$caId-root"

        sysOperations.mount(pkiPath, mount)

        return vaultTemplate.doWithSession {
            val response = it.exchange(
                "{pkiPath}/root/generate/internal",
                HttpMethod.POST,
                HttpEntity(
                    VaultGenRootRequest(
                        "$caId Root CA 1",
                        csrFields,
                        ttl,
                        request.keyType,
                        request.keyBits,
                    )
                ),
                VaultCreateCAResponse::class.java,
                mapOf("pkiPath" to pkiPath)
            )

            response.body!!.data
        }
    }

    fun createIntermediateCsr(
        caId: String,
        request: CreateCARequest
    ): VaultCsrResponse.Data {
        val csrFields = request.csrFields ?: CsrRequest()
        val builder = VaultMount.builder()
        val ttl = request.intermediateTTL ?: INTERMEDIATE_TTL

        builder.type("pki")
        // 1 month - 5 years validity for leaf certificates
        builder.config(
            mapOf(
                "max_lease_ttl" to ttl,
                "default_lease_ttl" to ttl
            )
        )
        builder.description("Intermediate CA for $caId")

        val mount = builder.build()
        val sysOperations = vaultTemplate.opsForSys()
        val pkiPath = "$pkiManagedPath/$caId-int-01"

        sysOperations.mount(pkiPath, mount)

        return vaultTemplate.doWithSession {
            val response = it.exchange(
                "{pkiPath}/intermediate/generate/internal",
                HttpMethod.POST,
                HttpEntity(
                    VaultGenIntermediateRequest(
                        "$caId Intermediate CA 01",
                        csrFields,
                        request.keyType,
                        request.keyBits,
                    )
                ),
                VaultCsrResponse::class.java,
                mapOf("pkiPath" to pkiPath)
            )

            response.body!!.data
        }
    }

    fun setSignedIntermediateCA(
        caId: String,
        certChain: String
    ) {
        val pkiPath = "$pkiManagedPath/$caId-int-01"

        vaultTemplate.doWithSession {
            it.exchange(
                "{pkiPath}/intermediate/set-signed",
                HttpMethod.POST,
                HttpEntity(VaultSetSignedRequest(certChain)),
                String::class.java,
                mapOf("pkiPath" to pkiPath)
            )
        }
    }

    fun signIntermediateCA(
        caId: String,
        csr: String,
        ttl: String
    ): VaultCreateCAResponse.Data {
        val pkiPath = "$pkiManagedPath/$caId-root"

        val data = vaultTemplate.doWithSession {
            val response = it.exchange(
                "{pkiPath}/root/sign-intermediate",
                HttpMethod.POST,
                HttpEntity(VaultSignIntermediateRequest(csr, ttl)),
                VaultCreateCAResponse::class.java,
                mapOf("pkiPath" to pkiPath)
            )

            if (response.statusCode.is4xxClientError) {
                throw Exception("bad request from Vault server: ${response.body}")
            }

            response.body!!.data
        }

        setSignedIntermediateCA(caId, "${data.certificate}\n${data.issuingCa}")
        return data
    }

    @ExceptionHandler(VaultException::class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    fun handleVaultException(e: VaultException): String =
        when (val rootCause = e.rootCause) {
            is HttpClientErrorException ->
                when {
                    rootCause.statusCode.is4xxClientError ->
                        "bad request from Vault server: ${rootCause.message}"
                    else -> throw e
                }
            else -> throw e
        }
}
