package eu.devity.vaultconnector.controller

import eu.devity.vaultconnector.serializable.*
import eu.devity.vaultconnector.serializable.vault.VaultCertificateResponse
import eu.devity.vaultconnector.serializable.vault.VaultRevokeCertificateRequest
import eu.devity.vaultconnector.serializable.vault.VaultRevokeCertificateResponse
import org.bouncycastle.cert.X509CertificateHolder
import org.bouncycastle.openssl.PEMParser
import org.slf4j.Logger
import org.slf4j.LoggerFactory
import org.springframework.http.HttpEntity
import org.springframework.http.HttpMethod
import org.springframework.http.HttpStatus
import org.springframework.vault.VaultException
import org.springframework.vault.core.VaultTemplate
import org.springframework.web.bind.annotation.*
import org.springframework.web.client.HttpClientErrorException
import org.springframework.web.util.UriUtils
import java.io.StringReader

@RestController
@RequestMapping("/certificates")
class CertificateController(val vaultTemplate: VaultTemplate) {

    var logger: Logger = LoggerFactory.getLogger(this::class.java)

    @PostMapping
    @ResponseStatus(HttpStatus.CREATED)
    fun createCertificate(
        @RequestBody request: CreateCertificateRequest,
    ): CreateCertificateResponse =
        if (request.requestType == RequestType.ISSUE) issueCertificate(
            request.dnAndExt!!,
            request.pkiPath,
            request.roleName
        ) else signCertificateRequest(
            request.csr!!,
            request.pkiPath,
            request.roleName
        )

    @PostMapping(value = ["revoke", "revoke/{serialNumber}"])
    @ResponseStatus(HttpStatus.OK)
    fun revokeCertificate(
        @PathVariable(required = false) serialNumber: String?,
        @RequestBody request: RevokeCertificateRequest,
    ): VaultRevokeCertificateResponse.Data {
        val sn = if (serialNumber == null) {
            val pem = PEMParser(StringReader(request.certificate!!))
            val certificate = pem.readObject()
            if (certificate !is X509CertificateHolder) {
                throw IllegalArgumentException("expect either a valid x509 certificate or a serial number")
            }
            "%x".format(certificate.serialNumber).chunked(2).joinToString(":")
        } else {
            UriUtils.decode(serialNumber, "UTF-8").replace('-', ':')
        }

        return vaultTemplate.doWithSession {
            val response = it.exchange(
                "{pkiPath}/revoke",
                HttpMethod.POST,
                HttpEntity(VaultRevokeCertificateRequest(sn)),
                VaultRevokeCertificateResponse::class.java,
                mapOf("pkiPath" to request.pkiPath)
            )
            if (response.statusCode.value() == 204)
                VaultRevokeCertificateResponse.Data(null, true)
            else
                response.body!!.data
        }
    }

    fun issueCertificate(
        dnAndExt: GenerateCertificateRequest,
        pkiPath: String,
        roleName: String,
    ): CreateCertificateResponse {
        return vaultTemplate.doWithSession {
            val response = it.exchange(
                "{pkiPath}/issue/{roleName}",
                HttpMethod.POST,
                HttpEntity(dnAndExt),
                VaultCertificateResponse::class.java,
                mapOf("pkiPath" to pkiPath, "roleName" to roleName)
            )

            val data = response.body!!.data
            CreateCertificateResponse(
                data.serialNumber,
                data.certificate,
                data.issuingCa,
                data.privateKey!!,
                data.caChain
            )
        }
    }

    fun signCertificateRequest(
        csr: String,
        pkiPath: String,
        roleName: String,
    ): CreateCertificateResponse {
        return vaultTemplate.doWithSession {
            val response = it.exchange(
                "{pkiPath}/sign/{roleName}",
                HttpMethod.POST,
                HttpEntity(SignCertificateRequest(csr, "not matter")),
                VaultCertificateResponse::class.java,
                mapOf("pkiPath" to pkiPath, "roleName" to roleName)
            )

            val data = response.body!!.data
            CreateCertificateResponse(
                data.serialNumber,
                data.certificate,
                data.issuingCa,
                null,
                data.caChain
            )
        }
    }

    @ExceptionHandler(VaultException::class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    fun handleVaultException(e: VaultException): String =
        when (val rootCause = e.rootCause) {
            is HttpClientErrorException ->
                when {
                    rootCause.statusCode.is4xxClientError -> {
                        logger.warn(rootCause.toString())
                        logger.error(e.stackTraceToString())
                        "bad request from Vault server: ${rootCause.message}"
                    }
                    else -> throw e
                }
            else -> throw e
        }
}
